from typing import Dict, List, Optional

import requests
from requests import HTTPError
from urllib3.util import retry

from databricks.rag_eval import env_vars, session
from databricks.rag_eval.clients.databricks_api_client import DatabricksAPIClient
from databricks.rag_eval.clients.llmjudge import proto_serde
from databricks.rag_eval.config import example_config
from databricks.rag_eval.config.assessment_config import (
    AssessmentType,
)
from databricks.rag_eval.evaluation import entities

SESSION_ID_HEADER = "eval-session-id"
BATCH_SIZE_HEADER = "eval-session-batch-size"


class LlmJudgeClient(DatabricksAPIClient):
    """
    Client to interact with the LLM judge service (/chat-assessments).
    """

    def __init__(
        self,
        api_url: str,
        api_token: str,
        session_metadata: Optional[session.SessionMetadata] = None,
    ):
        super().__init__(
            api_url=api_url,
            api_token=api_token,
            version="2.0",
        )
        self.extra_headers = self._construct_request_headers_from_session_metadata(
            session_metadata
        )
        self.proto_serde = proto_serde.ChatAssessmentProtoSerde()

    def get_assessment(
        self,
        eval_item: entities.EvalItem,
        assessment_name: str,
        assessment_type: AssessmentType,
        examples: List[example_config.AssessmentExample],
        domain_instructions: Optional[str],
    ) -> List[entities.AssessmentResult]:
        """
        Retrieves the assessment results from the LLM judge service for the given eval item and requested assessment
        """
        request_json = self.proto_serde.construct_assessment_request_json(
            eval_item, assessment_name, examples, domain_instructions
        )

        retries = retry.Retry(
            total=env_vars.RAG_EVAL_LLM_JUDGE_MAX_RETRIES.get(),
            backoff_factor=env_vars.RAG_EVAL_LLM_JUDGE_BACKOFF_FACTOR.get(),
            status_forcelist=[429, 500, 502, 503, 504],
            backoff_jitter=env_vars.RAG_EVAL_LLM_JUDGE_BACKOFF_JITTER.get(),
            allowed_methods=frozenset(
                ["GET", "POST"]
            ),  # by default, it doesn't retry on POST
        )

        with self.get_request_session(retries) as session:
            session.headers.update(self.extra_headers)
            resp = session.post(
                self.get_method_url("/rag-studio/chat-assessments"),
                json=request_json,
                auth=self.get_auth(),
                headers=self.extra_headers,
            )

        if resp.status_code == requests.codes.ok:
            return self.proto_serde.construct_assessment_result(
                resp.json(), assessment_name
            )
        else:
            try:
                resp.raise_for_status()
            except HTTPError as e:
                return self.proto_serde.construct_assessment_error_result(
                    assessment_name,
                    assessment_type,
                    resp.status_code,
                    e,
                )

    @classmethod
    def _construct_request_headers_from_session_metadata(
        cls, session_metadata: Optional[session.SessionMetadata]
    ) -> Dict[str, str]:
        """Constructs the request headers from the session metadata."""
        if session_metadata is None:
            return {}
        headers = {
            SESSION_ID_HEADER: session_metadata.session_id,
        }
        if session_metadata.session_batch_size is not None:
            headers[BATCH_SIZE_HEADER] = str(session_metadata.session_batch_size)

        return headers
