"""
Introduces main Context class and the framework to specify different specialized
contexts.
"""

from __future__ import annotations

import functools
import inspect
import logging
import threading
import uuid
from abc import ABC, abstractmethod

import mlflow
from mlflow.utils import databricks_utils

from databricks.rag_eval import session
from databricks.rag_eval.clients import llmjudge
from databricks.rag_eval.clients.llmjudge import llm_judge_client

_logger = logging.getLogger(__name__)


class Context(ABC):
    """
    Abstract class for execution context.
    Context is stateful and is used to store information related to the current eval session.
    """

    @abstractmethod
    def clear(self) -> None:
        """
        Clears the context and frees up any resources as needed.
        """
        pass

    @abstractmethod
    def display_html(self, html: str) -> None:
        """
        Displays HTML in the current execution context.
        """
        pass

    @abstractmethod
    def build_llm_judge_client(self) -> llmjudge.LlmJudgeClient:
        """
        Build a LLM Judge client for the current eval session.
        """
        pass

    @property
    @abstractmethod
    def session_metadata(self) -> session.SessionMetadata:
        """
        Gets the metadata for the current eval session.
        """
        pass


class NoneContext(Context):
    """
    A context that does nothing.
    """

    def clear(self) -> None:
        pass

    def display_html(self, html: str) -> None:
        raise AssertionError("Context is not set")

    def build_llm_judge_client(self) -> llmjudge.LlmJudgeClient:
        raise AssertionError("Context is not set")

    @property
    def session_metadata(self) -> session.SessionMetadata:
        raise AssertionError("Context is not set")


class RealContext(Context):
    """
    Context for eval execution.

    NOTE: This class is not covered by unit tests and is meant to be tested through
    smoke tests that run this code on an actual Databricks cluster.
    """

    @classmethod
    def _get_dbutils(cls):
        """
        Returns an instance of dbutils.
        """
        try:
            from databricks.sdk.runtime import dbutils

            return dbutils
        except ImportError:
            import IPython

            dbutils = IPython.get_ipython().user_ns["dbutils"]
        return dbutils

    def __init__(self):
        self._dbutils = self._get_dbutils()

        self._session_metadata = session.SessionMetadata(
            session_id=str(uuid.uuid4()),
            session_batch_size=None,
        )

        # Set MLflow model registry to Unity Catalog
        mlflow.set_registry_uri("databricks-uc")

    def clear(self) -> None:
        pass

    def display_html(self, html) -> None:
        # pylint: disable=protected-access
        self._dbutils.notebook.displayHTML(html)

    def build_llm_judge_client(self) -> llm_judge_client.LlmJudgeClient:
        host_creds = databricks_utils.get_databricks_host_creds()
        api_url = host_creds.host
        api_token = host_creds.token
        return llm_judge_client.LlmJudgeClient(
            api_url=api_url, api_token=api_token, session_metadata=self.session_metadata
        )

    @property
    def session_metadata(self) -> session.SessionMetadata:
        return self._session_metadata


# We use a thread-local storage to store the context for the current thread so that we allow multi-thread
# execution of the eval APIs.
contexts = threading.local()
CONTEXT_KEY = "context"


def context_is_active() -> bool:
    """
    Check if a context is active in the current thread.
    """
    return not isinstance(current_context(), NoneContext)


def current_context() -> Context:
    """
    Get the current context in the current thread.
    """
    return getattr(contexts, CONTEXT_KEY, NoneContext())


def set_context(context_impl: Context) -> None:
    """
    Set the current context in the current thread.
    """
    setattr(contexts, CONTEXT_KEY, context_impl)


def clear_context() -> None:
    """
    Clear the current context in the current thread.
    """
    current_context().clear()
    set_context(NoneContext())


def build_context() -> Context:
    """
    Build a new context.
    """
    return RealContext()


def wrap_with_context(func):
    """
    Decorator for wrapping all eval APIs with setup and closure logic.

    The wrapper will set up an active DatabricksContext if there isn't one already,
    and will also instrument the API call to add usage logging by timing it and
    recording the parameters and errors (if any).

    :param func: eval function to wrap
    :return: return value of func
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        set_context(build_context())

        error = None
        result = None

        parameters = inspect.signature(func).parameters
        # Get all the parameters with default values from the method signature.
        # If a parameter does not have default value, it will not be included in the `full_kwargs`.
        full_kwargs = {
            param_name: parameters[param_name].default
            for param_name in parameters
            if parameters[param_name].default != inspect.Parameter.empty
        }
        # Merge the parameters default values with the values passed in by the user.
        full_kwargs.update(kwargs)

        # Do any preprocessing of the args here

        try:
            result = func(*args, **full_kwargs)
        except Exception as e:  # pylint: disable=broad-except
            error = e
        finally:
            # Clear the context
            clear_context()
            # Raise the original error if there was one, otherwise return
            if error is not None:
                raise error
            else:
                return result  # pylint: disable=lost-exception

    return wrapper
