import json
import tempfile
import textwrap
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import mlflow
import pandas as pd
from mlflow import models as mlflow_models
from mlflow.models import evaluation as mlflow_evaluation
from mlflow.models.evaluation import artifacts as mlflow_artifacts
from mlflow.utils import mlflow_tags

from databricks.rag_eval import constants, context, schemas
from databricks.rag_eval.config import evaluation_config
from databricks.rag_eval.evaluation import entities, harness
from databricks.rag_eval.judges import builtin_judge
from databricks.rag_eval.mlflow import datasets, mlflow_log_metrics
from databricks.rag_eval.utils import error_utils, workspace_url_resolver


def _log_pandas_df_artifact(pandas_df, artifact_name):
    """
    Logs a pandas DataFrame as a JSON artifact, then returns an EvaluationArtifact object.
    """
    mlflow_artifact_name = f"{artifact_name}.json"
    mlflow.log_table(pandas_df, mlflow_artifact_name)
    return mlflow_artifacts.JsonEvaluationArtifact(
        uri=mlflow.get_artifact_uri(mlflow_artifact_name),
    )


def generate_summary(results_df: pd.DataFrame) -> Tuple[str, str]:
    """
    Generates a short summary of the evaluation results suitable for printing

    :param results_df: DataFrame containing evaluation results
    :return: Tuple containing the summary in HTML and text
    """

    correct_rating_col = schemas.get_response_llm_rating_col_name(
        builtin_judge.CORRECTNESS.assessment_name
    )
    correct_error_col = schemas.get_response_llm_error_message_col_name(
        builtin_judge.CORRECTNESS.assessment_name
    )
    total_count = len(results_df)
    if correct_rating_col in results_df.columns:
        correct_count = (
            results_df[correct_rating_col] == entities.CategoricalRating.YES
        ).sum()
        total_rated_count = results_df[correct_rating_col].notnull().sum()
    else:
        correct_count = 0
        total_rated_count = 0

    if correct_error_col in results_df.columns:
        error_count = results_df[correct_error_col].notnull().sum()
    else:
        error_count = 0

    missing_count = total_count - total_rated_count - error_count

    # Let's cover corner cases first
    if missing_count == total_count:
        # Print nothing -- we already print a warning in this case
        correct_summary_text_html = ""
        correct_summary_text = ""
    else:
        msg = f"{correct_count}/{total_count} correct responses"
        aux_msgs = []
        if error_count > 0:
            aux_msgs.append(f"{error_count} error{'s' if error_count > 1 else ''}")
        if missing_count > 0:
            aux_msgs.append(f"{missing_count} missing ground truth")

        if len(aux_msgs) > 0:
            msg += f" ({', '.join(aux_msgs)})"

        more_results_msg = (
            "You can access more information about answer correctness and other judgements in the "
            "evaluation results."
        )

        correct_summary_text_html = f"<b>{msg}</b>&nbsp;&nbsp;{more_results_msg}"
        correct_summary_text = f"{msg} -- {more_results_msg}"

    return correct_summary_text_html, correct_summary_text


def _emit_usage_instructions(run_id: str, results_df: pd.DataFrame):
    """
    Displays instructions to the user on what to do after running `mlflow.evaluate`.
    """

    summary_html, summary_text = generate_summary(results_df)

    run = mlflow.get_run(run_id)
    if mlflow_tags.MLFLOW_DATABRICKS_WORKSPACE_URL in run.data.tags:
        # Include Databricks URLs in the displayed message.
        workspace_url = run.data.tags[mlflow_tags.MLFLOW_DATABRICKS_WORKSPACE_URL]
        resolver = workspace_url_resolver.WorkspaceUrlResolver(workspace_url)
        evaluation_url = resolver.resolve_url_for_mlflow_experiment_eval_view(run.info)
        run_url = resolver.resolve_url_for_mlflow_run(run.info)

        html_header = (
            "Evaluation completed:&nbsp;" if summary_html else "Evaluation completed."
        )
        context.current_context().display_html(
            textwrap.dedent(
                f"""
            <b>{html_header}</b>{summary_html}<br><br>
            Aggregate metrics are available in <code>mlflow.evaluate(...).metrics</code>.<br>
            Per-row evaluation results are available in <code>mlflow.evaluate(...).tables['eval_results']</code>.
            <br><br>

            Metrics and evaluation results are logged to MLflow, and can be viewed from the 
            <a href='{run_url}'>MLflow run page</a>.
            To compare evaluation results across runs, view the <a href='{evaluation_url}'>Evaluation tab</a> of 
            the experiement.
            """
            )
        )
    else:
        text_header = (
            "Evaluation completed: " if summary_text else "Evaluation completed."
        )
        print(
            f"""{text_header} {summary_text}
            
Aggregate metrics are available in mlflow.evaluate(...).metrics.
Per-row evaluation results are available in mlflow.evaluate(...).tables['eval_results'].

Metrics and evaluation results are logged to MLflow, and can be viewed from the MLflow run page,
or the "Evaluation" tab of the Experiment page."""
        )


def _log_evaluation_input_artifacts(
    config: evaluation_config.EvaluationConfig, examples_df: Optional[pd.DataFrame]
):
    """
    Logs the configuration to MLflow, as well as the examples_df if provided.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        path = Path(temp_dir, "eval_config.json")
        config_str = json.dumps(config.to_dict())
        path.write_text(config_str)
        mlflow.log_artifact(path)

    if examples_df is not None:
        _log_pandas_df_artifact(examples_df, "eval_examples")


class DatabricksRagEvaluator(mlflow_evaluation.base.ModelEvaluator):
    def can_evaluate(self, *, model_type, evaluator_config, **kwargs) -> bool:
        """
        See parent class docstring.
        """
        return model_type == "databricks-rag"

    @context.wrap_with_context
    def evaluate(
        self,
        *,
        model_type,
        dataset,
        run_id,
        evaluator_config: Optional[Dict[str, Any]] = None,
        model=None,
        custom_metrics=None,
        extra_metrics=None,
        custom_artifacts=None,
        baseline_model=None,
        predictions=None,
        **kwargs,
    ):
        """
        Runs Databricks RAG evaluation on the provided dataset.

        The following arguments are supported:
        - model_type: Must be "databricks-rag".
        - dataset
        - run_id

        For more details, see parent class docstring.
        """
        try:
            if evaluator_config is None:
                evaluator_config = {}

            eval_dataset = datasets.RagEvaluationDataset.from_mlflow_dataset(dataset)
            eval_items = eval_dataset.eval_items
            # Set batch size to the context
            context.current_context().session_metadata.session_batch_size = len(
                eval_items
            )

            config = evaluation_config.EvaluationConfig.from_mlflow_evaluate_args(
                evaluator_config, extra_metrics
            )
            examples_df: pd.DataFrame = evaluator_config.get(
                constants.EVALUATOR_CONFIG_EXAMPLES_KEY_NAME, None
            )

            _log_evaluation_input_artifacts(config, examples_df)

            eval_results = harness.run(
                eval_items=eval_items, config=config, model=model
            )

            mlflow_metrics = mlflow_log_metrics.generate_mlflow_metrics(eval_results)
            mlflow.log_metrics(mlflow_metrics)

            eval_results_df = pd.DataFrame(
                [result.to_pd_series() for result in eval_results]
            )

            eval_results_artifact = _log_pandas_df_artifact(
                eval_results_df, "eval_results"
            )
            result = mlflow_models.EvaluationResult(
                metrics=mlflow_metrics,
                artifacts={"eval_results": eval_results_artifact},
            )
            _emit_usage_instructions(run_id, eval_results_df)
            return result
        except error_utils.ValidationError as e:
            # Scrub trace for user-facing validation errors
            raise error_utils.ValidationError(str(e)) from None
